@@ -7,8 +7,10 @@ Downloads, manages, and runs SpeakLeash models directly from Hugging Face.
 import os
 import json
 import shutil
+import signal
+import time
 from pathlib import Path
-from typing import Dict, List, Optional, Tuple, Any
+from typing import Dict, List, Optional, Tuple, Any, Callable
 from dataclasses import dataclass, asdict
 import logging
 
@@ -23,6 +25,75 @@ except ImportError:
 
 from .config import get_config, get_logger
 
+# Custom exceptions
+class ModelLoadingError(Exception):
+    """Base exception for model loading errors."""
+    pass
+
+class ModelLoadingTimeoutError(ModelLoadingError):
+    """Raised when model loading exceeds the timeout."""
+    pass
+
+def get_model_loading_timeout() -> int:
+    """Get the model loading timeout from environment or use default."""
+    return int(os.environ.get("BIELIK_LOAD_TIMEOUT", "10"))
+
+def is_debug_mode() -> bool:
+    """Check if debug mode is enabled via environment variable."""
+    return os.environ.get("BIELIK_DEBUG", "0").lower() in ("1", "true", "yes")
+
+def timeout_handler(signum, frame):
+    """Handle the timeout signal."""
+    raise ModelLoadingTimeoutError("Model loading timed out")
+
+def load_with_timeout(loader_func: Callable, timeout: int, **kwargs) -> Any:
+    """
+    Run a loader function with a timeout.
+    
+    Args:
+        loader_func: Function that loads and returns the model
+        timeout: Maximum time in seconds to wait
+        **kwargs: Arguments to pass to the loader function
+        
+    Returns:
+        The loaded model
+        
+    Raises:
+        ModelLoadingTimeoutError: If loading exceeds the timeout
+        ModelLoadingError: For other loading errors
+    """
+    logger = get_logger(__name__)
+    debug = is_debug_mode()
+    
+    if debug:
+        logger.setLevel(logging.DEBUG)
+        logger.debug(f"Starting model loading with timeout={timeout}s")
+    
+    # Set the signal handler and a timeout
+    original_handler = signal.signal(signal.SIGALRM, timeout_handler)
+    signal.alarm(timeout)
+    
+    try:
+        start_time = time.time()
+        result = loader_func(**kwargs)
+        load_time = time.time() - start_time
+        logger.info(f"Model loaded successfully in {load_time:.2f} seconds")
+        return result
+    except ModelLoadingTimeoutError:
+        logger.error(f"Model loading timed out after {timeout} seconds")
+        if debug:
+            logger.debug("Debug info:")
+            logger.debug(f"Loader function: {loader_func.__name__}")
+            logger.debug(f"Loader arguments: {kwargs}")
+        raise
+    except Exception as e:
+        logger.error(f"Error loading model: {str(e)}", exc_info=debug)
+        raise ModelLoadingError(f"Failed to load model: {str(e)}")
+    finally:
+        # Disable the alarm and restore original handler
+        signal.alarm(0)
+        signal.signal(signal.SIGALRM, original_handler)
+
 
 @dataclass
 class ModelInfo:
@@ -92,27 +163,85 @@ class SpeakLeashModelManager:
         # Initialize HF API
         self.hf_api = HfApi()
         
-        # Load existing model registry
-        self.registry = self._load_registry()
+        # Initialize empty registry - will be loaded on first use
+        self.registry = {}
+        self._models_initialized = False
         
-        self.logger.info(f"Model manager initialized with directory: {self.models_dir}")
+        self.logger.info(f"Model manager initialized (lazy loading enabled) with directory: {self.models_dir}")
+    
+    def initialize_models(self):
+        """
+        Initialize models (lazy loading).
+        This is called automatically when models are first needed.
+        """
+        if not self._models_initialized:
+            self.registry = self._load_registry()
+            self._models_initialized = True
+            self.logger.info("Models initialized")
+    
+    def _ensure_initialized(self):
+        """Ensure models are initialized before operations that need them."""
+        if not self._models_initialized:
+            self.initialize_models()
     
     def _load_registry(self) -> Dict[str, ModelInfo]:
-        """Load model registry from file."""
+        """
+        Load model registry from file with validation.
+        
+        Returns:
+            Dictionary mapping model names to ModelInfo objects
+        """
         if not self.registry_file.exists():
+            self.logger.info("No model registry found, starting with empty registry")
             return {}
         
+        self.logger.info(f"Loading model registry from {self.registry_file}")
+        
         try:
             with open(self.registry_file, 'r', encoding='utf-8') as f:
                 data = json.load(f)
             
-            registry = {}
-            for name, info_dict in data.items():
-                registry[name] = ModelInfo(**info_dict)
+            if not isinstance(data, dict):
+                self.logger.error(f"Invalid registry format, expected dict, got {type(data).__name__}")
+                return {}
             
+            registry = {}
+            valid_models = 0
+            
+            for name, info in data.items():
+                try:
+                    if not isinstance(info, dict):
+                        self.logger.warning(f"Skipping invalid model info for {name}: not a dictionary")
+                        continue
+                        
+                    # Validate required fields
+                    required_fields = ['name', 'repo_id', 'file_name', 'local_path']
+                    if not all(field in info for field in required_fields):
+                        self.logger.warning(f"Skipping model {name}: missing required fields")
+                        continue
+                        
+                    # Convert to ModelInfo
+                    model_info = ModelInfo(**info)
+                    
+                    # Verify the model file exists
+                    if not Path(model_info.local_path).exists():
+                        self.logger.warning(f"Model file not found: {model_info.local_path}")
+                        continue
+                        
+                    registry[name] = model_info
+                    valid_models += 1
+                    
+                except Exception as e:
+                    self.logger.error(f"Error loading model {name}: {str(e)}", exc_info=self.config.DEBUG)
+            
+            self.logger.info(f"Successfully loaded {valid_models} models from registry")
             return registry
+            
+        except json.JSONDecodeError as e:
+            self.logger.error(f"Failed to parse registry file {self.registry_file}: {e}")
+            return {}
         except Exception as e:
-            self.logger.error(f"Failed to load model registry: {e}")
+            self.logger.error(f"Unexpected error loading registry: {e}", exc_info=self.config.DEBUG)
             return {}
     
     def _save_registry(self):
@@ -128,19 +257,30 @@ class SpeakLeashModelManager:
         except Exception as e:
             self.logger.error(f"Failed to save model registry: {e}")
     
-    def list_available_models(self) -> Dict[str, Dict[str, Any]]:
+    def list_available_models(self) -> List[Dict[str, str]]:
         """List all available SpeakLeash models."""
-        return self.SPEAKLEASH_MODELS.copy()
+        self._ensure_initialized()
+        return [
+            {
+                "name": name,
+                "description": info["description"],
+                "parameters": info["parameters"],
+                "version": info["version"]
+            }
+            for name, info in self.SPEAKLEASH_MODELS.items()
+        ]
     
     def list_downloaded_models(self) -> Dict[str, ModelInfo]:
         """List all downloaded models."""
+        self._ensure_initialized()
+        
         # Verify that files still exist
         valid_registry = {}
         for name, model_info in self.registry.items():
-            if Path(model_info.local_path).exists():
+            if model_info and hasattr(model_info, 'local_path') and Path(model_info.local_path).exists():
                 valid_registry[name] = model_info
             else:
-                self.logger.warning(f"Model file missing: {model_info.local_path}")
+                self.logger.warning(f"Model file missing or invalid: {getattr(model_info, 'local_path', 'unknown')}")
         
         # Update registry if needed
         if len(valid_registry) != len(self.registry):
@@ -150,12 +290,9 @@ class SpeakLeashModelManager:
         return self.registry.copy()
     
     def is_model_downloaded(self, model_name: str) -> bool:
-        """Check if model is already downloaded."""
-        if model_name not in self.registry:
-            return False
-        
-        model_info = self.registry[model_name]
-        return Path(model_info.local_path).exists()
+        """Check if a model is downloaded."""
+        self._ensure_initialized()
+        return model_name in self.registry
     
     def get_model_files(self, repo_id: str) -> List[str]:
         """Get list of GGUF files in repository."""
@@ -293,8 +430,14 @@ class SpeakLeashModelManager:
             self.logger.error(f"Failed to delete model {model_name}: {e}")
             return False
     
+    def get_model_info(self, model_name: str) -> Optional[ModelInfo]:
+        """Get information about a downloaded model."""
+        self._ensure_initialized()
+        return self.registry.get(model_name)
+    
     def get_model_path(self, model_name: str) -> Optional[str]:
         """Get local path for a downloaded model."""
+        self._ensure_initialized()
         if not self.is_model_downloaded(model_name):
             return None
         
@@ -327,11 +470,16 @@ class LocalLlamaRunner:
     
     def __init__(self, model_path: str, **kwargs):
         """
-        Initialize local Llama runner.
+        Initialize local Llama runner with optimized loading.
         
         Args:
             model_path: Path to GGUF model file
             **kwargs: Additional arguments for Llama initialization
+            
+        Raises:
+            ImportError: If llama-cpp-python is not installed
+            ModelLoadingError: If model loading fails
+            ModelLoadingTimeoutError: If model loading times out
         """
         if not HAS_LLAMA_CPP:
             raise ImportError("llama-cpp-python is required for local model execution")
@@ -339,6 +487,7 @@ class LocalLlamaRunner:
         self.config = get_config()
         self.logger = get_logger(__name__)
         self.model_path = model_path
+        self.model = None
         
         # Default parameters
         default_params = {
@@ -349,21 +498,58 @@ class LocalLlamaRunner:
         }
         
         # Merge with user params
-        params = {**default_params, **kwargs}
+        self.params = {**default_params, **kwargs}
         
-        self.logger.info(f"Loading model from: {model_path}")
-        self.logger.info(f"Model parameters: {params}")
+        # Load the model with timeout
+        self._load_model()
+    
+    def _load_model(self):
+        """Load the model with timeout and error handling."""
+        self.logger.info(f"Loading model from: {self.model_path}")
+        self.logger.debug(f"Model parameters: {self.params}")
+        
+        # Define the loader function
+        def loader():
+            return Llama(self.model_path, **self.params)
         
         try:
-            self.model = Llama(model_path, **params)
-            self.logger.info("Model loaded successfully")
-        except Exception as e:
-            self.logger.error(f"Failed to load model: {e}")
+            # Try loading with GPU first if specified
+            if self.params.get("n_gpu_layers", 0) > 0:
+                try:
+                    self.model = load_with_timeout(
+                        loader,
+                        timeout=get_model_loading_timeout()
+                    )
+                    self.logger.info("Model loaded successfully with GPU acceleration")
+                    return
+                except (ModelLoadingError, Exception) as e:
+                    self.logger.warning(f"GPU loading failed: {str(e)}. Falling back to CPU...")
+                    # Fall back to CPU
+                    self.params["n_gpu_layers"] = 0
+            
+            # Load with CPU
+            self.model = load_with_timeout(
+                loader,
+                timeout=get_model_loading_timeout()
+            )
+            self.logger.info("Model loaded successfully on CPU")
+            
+        except ModelLoadingTimeoutError as e:
+            self.logger.error("Model loading timed out. The model file might be corrupted or too large.")
+            if is_debug_mode():
+                self.logger.debug("Debug information:")
+                self.logger.debug(f"Model path: {self.model_path}")
+                self.logger.debug(f"File exists: {os.path.exists(self.model_path)}")
+                if os.path.exists(self.model_path):
+                    self.logger.debug(f"File size: {os.path.getsize(self.model_path) / (1024*1024):.2f} MB")
             raise
+        except Exception as e:
+            self.logger.error(f"Failed to load model: {str(e)}", exc_info=is_debug_mode())
+            raise ModelLoadingError(f"Failed to load model: {str(e)}")
     
     def chat(self, messages: List[Dict[str, str]], **kwargs) -> str:
         """
-        Generate chat response.
+        Generate chat response with progress tracking.
         
         Args:
             messages: List of message dicts with 'role' and 'content'
@@ -372,6 +558,9 @@ class LocalLlamaRunner:
         Returns:
             Generated response text
         """
+        import sys
+        import time
+        
         # Convert messages to prompt
         prompt = self._messages_to_prompt(messages)
         
@@ -381,59 +570,4 @@ class LocalLlamaRunner:
             "temperature": 0.7,
             "top_p": 0.9,
             "echo": False,
-            "stop": ["</s>", "<|im_end|>", "<|endoftext|>"]
-        }
-        
-        # Merge with user params
-        params = {**default_params, **kwargs}
-        
-        try:
-            self.logger.debug(f"Generating response with params: {params}")
-            
-            result = self.model(prompt, **params)
-            response = result['choices'][0]['text'].strip()
-            
-            self.logger.debug(f"Generated response length: {len(response)} characters")
-            return response
-            
-        except Exception as e:
-            self.logger.error(f"Failed to generate response: {e}")
-            return f"[LOCAL MODEL ERROR] {e}"
-    
-    def _messages_to_prompt(self, messages: List[Dict[str, str]]) -> str:
-        """Convert OpenAI-style messages to prompt format."""
-        # Simple prompt format - can be enhanced for specific models
-        prompt_parts = []
-        
-        for message in messages:
-            role = message.get('role', 'user')
-            content = message.get('content', '')
-            
-            if role == 'system':
-                prompt_parts.append(f"System: {content}")
-            elif role == 'user':
-                prompt_parts.append(f"User: {content}")
-            elif role == 'assistant':
-                prompt_parts.append(f"Assistant: {content}")
-        
-        # Add final prompt for assistant response
-        prompt_parts.append("Assistant:")
-        
-        return "\n".join(prompt_parts)
-    
-    def __del__(self):
-        """Cleanup model when object is destroyed."""
-        if hasattr(self, 'model'):
-            del self.model
-
-
-# Global model manager instance
-_model_manager: Optional[SpeakLeashModelManager] = None
-
-
-def get_model_manager() -> SpeakLeashModelManager:
-    """Get global model manager instance."""
-    global _model_manager
-    if _model_manager is None:
-        _model_manager = SpeakLeashModelManager()
-    return _model_manager
+            "stop": ["</s>", "
